package ParseArgs;

# ParseArgs.pm
#
# Perl5 package for table-driven argument parsing, somewhat like Tk's
# ParseArgv.  There's only one routine in this package that you *have*
# to call, and that's `Parse'; it takes two arguments: a reference to
# a table of options, and a reference to the list of command-line
# arguments that you want parsed.
#
# The argument table (first parameter to `Parse') consists of one
# element per valid command-line option; each element should be a
# reference to a list of the form:
#
#    ( option, type, num, value, help-string )
#
# where:
#
#  * option is the command-line option itself, *including* the option
#    prefix (eg. "-verbose" or "--long-gnu-style-option").  For
#    boolean options, you can also specify the "negative" option here,
#    as in "-execute|-noexecute" or "-quiet|-verbose" (this is
#    explained further below)
#    
#  * type specifies the behaviour to be taken when this option is
#    encountered.  type interacts in a rather complicated way with num
#    and value, as follows:
#      * if type is "integer", "float", or "string" then ParseArgs
#        expects the option to be followed by one or more values of the
#        given type.  (The validity of the values are checked with
#        fairly straightforward regular expressions.)  num specifies
#        just how many values should be expected to follow the option,
#        and value specifies where the values should be put.  (value
#        must be a reference to a scalar if num is 1, and to an array if
#        num is > 1.)  For instance, if you specify ("-num", "integer",
#        3, \@nums, "") as one of the elements of the argument table, then
#        "-num 1 2 3" on the command line will result in @nums (in
#        ParseArg's calling package) being set to (1,2,3).  However,
#        "-num 1 2 x" will result in an error from ParseArgs.
#      * if type is "constant", then num is just a single value that
#        will be assigned to whatever `value' references.  (I.e.,
#        `value' should be a scalar reference, and the scalar it
#        references will get the value in `num' when `option' is
#        encountered on the command line.)  Hence, a table element of
#        ("-thingy", "constant", 37, \$thingy, "") will result in $thingy
#        being assigned 37 if "-thingy" is seen on the command line.
#      * if type is "boolean", then *two* valid options are implied;
#        there is a "negated version" of each option.  Currently, you
#        must specify the negative version in the `option' string,
#        eg. "-execute|-noexecute".  Apart from that, the two options
#        act like a constant option where `num' is either 1 (for the
#        positive form) or 0 (for the negated form).  Thus, `num' is
#        ignored by boolean options, since it is either 1 or 0.
#      * if type is "call", then value should just be a reference to a
#        subroutine.  (It could also be a string containing name of a
#        subroutine in the calling package, but that form is
#        deprecated.)  When `option' is seen on the command line, the
#        referenced subroutine will be called with two arguments: the
#        option that was seen (because you could specify the same
#        subroutine for more than one option), and a reference to a
#        list containing all the arguments after the option.  The
#        subroutine should remove any arguments that it uses from this
#        list (unless they should be processed twice, for some
#        perverse reason).  Also, it should return 1 on success or 0
#        on error (which will cause &Parse to immediately return 0).
#      * if type is "eval", then value should just be a string
#        containing a little chunk of code to evaluate when the option
#        is seen.  The code will be evaluated in the package from which
#        Parse was called, so you don't need to explicitly set the
#        package of any variables that are set (or whatever) in the code.
#      * if type is "copy", then *either* n will be copied to the
#        location referenced by value, or (if n is undefined), the
#        option string itself will be copied.  In the first case, this
#        is just like a "constant" option; however, if you give a
#        table element ("-foo", "copy", undef, \$foo), then $foo will
#        be set to "-foo" when "-foo" is seen on the command line.
#        This can be useful for accepting command line options that
#        are simply passed to child programs without further
#        inspection.
#      * if type is "copyarray", then n and value must both be
#        references to as arrays; the array referenced by n is copied
#        to that referenced by value.
#      * if type is "copyhash", it acts just like "copyarray", except
#        n and value must be references to associative arrays.
#
#   * help-string is a string explaining the option.  Don't bother
#     formatting this, as it's done internally for you.
# 
# The second argument to Parse is usually just a reference to ARGV.
# When Parse is finished, then all options and their arguments will
# have been removed, leaving just the "required" command-line
# arguments.  Also, any code called by "call" or evaluated by "eval"
# options can modify the argument list, and these changes will be
# reflected after Parse is finished.
# 
# If there are any errors in the command line (unknown option, or a
# bad argument to a typed option, eg. giving a floating point number
# where an integer was expected), then Parse will print out a
# reasonably complete error message and return 0.  Most of the
# time, this message will be sufficient that the caller won't have to
# do anything more -- you can just die silently.
#
# by Greg Ward 95/07/06 - 95/07/09

# $Id: ParseArgs.pm,v 1.13 1996/02/12 20:18:50 greg Exp $

# --------------------------------------------------------------------
# Copyright (c) 1995 Greg Ward, McConnell Brain Imaging Centre,
# Montreal Neurological Institute, McGill University.  Permission to
# use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appear in all copies.  The
# author and McGill University make no representations about the
# suitability of this software for any purpose.  It is provided "as
# is" without express or implied warranty.
# --------------------------------------------------------------------

# -------------------------------------------------------------------- #
# Private global variables                                             #
# -------------------------------------------------------------------- #

# regexp for C floating point numbers adapted from the grammar for
# ANSI C that comes with pccts 1.1 (ack. to Terence Parr)

%Patterns = ("integer"=>"\\d+",
             "float"  =>"[+-]?([1-9]\\d*(\\.\\d*)?|0?\\.\\d+)([Ee][+-]?\\d+)?",
             "string" =>".*");

%KnownTypes = ( "constant", 1, 
                "string",   1,
                "integer",  1,
                "float",    1,
                "boolean",  1,
                "call",     1,
                "eval",     1,
                "copy",     1,
                "copyarray",1,
                "copyhash", 1,
                "section",  1);

%TypeDescriptions = ("integer" => "integer", 
                     "float"   => "floating-point number",
                     "string"  => "string");

@OptionPrefixes = ("-");
$OptionTerminator = "--";
%OptionNegatives = ();
$AllowTermination = 1;

# -------------------------------------------------------------------- #
# Public subroutines used to set options before calling Parse.         #
# -------------------------------------------------------------------- #

sub SetHelpText
{
   $LongHelp = shift;
   $Usage = shift;
}

sub SetPrefixes
{
   @OptionPrefixes = @_;
}

sub SetTerminator
{
   $OptionTerminator = shift;
   $AllowTermination = 1;
}

sub UnsetTerminator
{
   $AllowTermination = 0;
}

sub AddType
{
   local ($type, $pattern) = @_;
   $KnownTypes{$type} = 1;
   $Patterns{$type} = $pattern;
}

sub GetPattern
{
   local ($type) = @_;
   $Patterns{$type};
}

sub SetOptionNegative
{
   local ($opt, $neg) = @_;
   $OptionNegatives{$opt} = $neg;
}


# -------------------------------------------------------------------- #
# Private subroutines called by Parse (and each other)                 #
# -------------------------------------------------------------------- #

sub printhelp
{
   die "ParseArgs: internal error, wrong number of input args to &printhelp"
      if (scalar (@_) != 1);
   local ($argtable) = @_;
   local ($maxlength, $opt, $curlength, $format, $breakers);

   $maxlength = 0;
   foreach $opt (@$argtable)
   {
      ($option,$type) = @$opt;
      next if $type eq "section";
      ($option) = &split_option ($opt) if $type eq "boolean";
      $curlength = length ($option);
      $maxlength = $curlength if ($curlength > $maxlength);
   }

   # We need to construct and eval code that looks something like this:
   #    format =
   #    @<<<<<<<<<<<<<<<  ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
   # $option,        $help
   # ~~                   ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
   #                 $help
   # .
   # in order to nicely print out the help.  Can't hardcode a format, 
   # though, because we don't know until now how much space to allocate
   # for the option (ie. $maxlength).

   $breakers = $:;
   $: = " \n";
   chop ($numcols = `tput cols`);
   $numcols = 80 if $? || !$numcols;

   $textlength = $numcols - 3 - $maxlength - 2;
   $format = "format = \n" .
      "   @" . ("<" x ($maxlength-1)) . "  ^" . ("<" x ($textlength-1)) . "\n".
      "\$option, \$help\n" .
      "~~ " . (" " x $maxlength) . "  ^" . ("<" x ($textlength-1)) . "\n" .
      "\$help\n.";
   eval $format;
   die "INTERNAL ERROR with \$format: $@" if $@;

   print $LongHelp . "\n" if defined $LongHelp;
   print "Summary of options:\n";
   foreach $opt (@$argtable)
   {
      ($option,$type,$help) = @{$opt}[0,1,4];

      if ($type eq "section")
      {
	 printf "\n-- %s %s\n", $option, "-" x ($numcols-4-length($option));
      }
      else
      {
	 ($option) = &split_option ($opt) if $type eq "boolean";
	 write;
      }
   }
   $: = $breakers;
   print "\n" . $Usage if defined $Usage;
}


sub scan_table
{
   local ($argtable, $arghash) = @_;
   local ($opt, $option, $type, $value);

   for $i (0 .. $#$argtable)
   {
      $opt = $argtable->[$i];
      ($option, $type, $value) = @$opt;
      unless ($KnownTypes{$type})
      {
	 warn "ParseArgs: unknown type ($type) supplied for option $option\n";
	 return 0;
      }
      if ($type eq "boolean")
      {
         ($pos,$neg) = &split_option($opt);
	 $arghash->{$pos} = $i;
         $arghash->{$neg} = $i if defined $neg;
      }
      elsif ($type ne "section")
      {
	 $arghash->{$option} = $i;
      }
   }
   return 1;
}


sub match_abbreviation
{
   my ($s, $words, $err_format) = @_;
   my ($match);

   foreach $word (@$words)
   {
      next unless ($s eq substr ($word, 0, length ($s)));

      # We have an exact match, so return it now

      return $word if ($s eq $word);

      # We have an approx. match, and already had one before

      if ($match)
      {				
	 warn sprintf ("$err_format\n", "ambiguous", $s);
	 return 0;
      }

      $match = $word;
   }
   warn sprintf ("$err_format\n", "unknown", $s) unless $match;
   $match;
}
   


sub check_value
{
   local ($option, $val, $n, $type) = @_;
   local ($article, $desc);

   unless (defined $val && $val =~ /^$Patterns{$type}$/)
   {
      $article = ($type =~ /^[aeiou]/) ? "an" : "a";
      $tdesc = $TypeDescriptions{$type};
      $desc = ($n > 1) ? 
	 "$n ${tdesc}s" : 
	 "$article $tdesc";
      warn "$option option must be followed by $desc\n";
      return 0;
   }      
}


sub split_option
{
   local ($opt_desc) = @_;

#   ($option, $neg) = @$opt_desc[0,2];
   $option = $opt_desc->[0];
   @options = split ('\|', $option);

   if (@options == 2)
   {
      return @options;
   }
   elsif (defined $OptionNegatives{$option})
   {
      return ($option, $OptionNegatives{$option});
   }
   else
   {
      return ($option);
   }
}



# -------------------------------------------------------------------- #
# The main public subroutine: Parse                                    #
# -------------------------------------------------------------------- #

sub Parse
{
   local ($argtbl, $arglist) = @_;
   local (@new_arglist, %argpos, $arg, $pos, $opt_ref);
   local ($prefix_re, $help_option, @option_list);

   &scan_table ($argtbl, \%argpos) || return 0;
   $help_option = $OptionPrefixes[0] . "help";
   $prefix_re = join ("|", @OptionPrefixes);

   foreach $opt_ref (@$argtbl)
   {
      push (@option_list, &split_option ($opt_ref))
	 unless $opt_ref->[1] eq "section";
   }

   while ($arg = shift @$arglist)
   {
#     print "arg: $arg\n";
      last if ($AllowTermination && $arg eq $OptionTerminator);
      if ($arg !~ /^($prefix_re)/)
      {
         push (@new_arglist, $arg);
         next;
      }

      if ($arg eq $help_option)
      {
         &printhelp ($argtbl);
         @$arglist = @new_arglist;
         return 0;
      }

      $arg = &match_abbreviation ($arg, \@option_list, "%s option: %s");
      if (! $arg)
      {
         @$arglist = @new_arglist;
         return 0;
      }

      $pos = $argpos{$arg};
      die "INTERNAL ERROR: didn't find arg in arg hash even " .
	  "after resolving abbreviation"
	     unless defined $pos;

      local ($option, $type, $n, $value) = @{$argtbl->[$pos]};

      if ($type eq "constant")
      {
         $$value = $n;
      }
      elsif ($type eq "boolean")
      {
         ($pos,$neg) = &split_option($argtbl->[$pos]);
         if ($arg eq $pos)
	 {
	    $$value = 1;
	 }
         elsif ($arg eq $neg)
	 {
	    $$value = 0;
	 }
         else
         {
            die "ParseArgs: INTERNAL ERROR: argument position hash corrupt";
         }
      }
      elsif ($type eq "call")
      {
	 if (ref $value eq "CODE")
	 {
	    return 0 unless &$value ($arg, $arglist);
	 }
	 else
	 {
	    my ($up_pkg, $result);

	    warn "warning: ParseArgs: using obsolete form of `call' type\n";

	    $up_pkg = (caller)[0];
	    $evalstr = <<END;
package $up_pkg;
&${value} (\$ParseArgs::arg, \$ParseArgs::arglist);
END
            $result = eval $evalstr;

            if ($@)
	    {
	       warn "Error calling $value: $@\n";
	       return 0;
	    }
   	    return 0 unless $result;
         }
      }
      elsif ($type eq "eval")
      {
	 my ($up_pkg) = (caller)[0];
         eval "package $up_pkg; $value";

         if ($@)		# any error string set?
	 {
	    warn "Error evaluating \"$value\": $@\n";
	    return 0;
	 }
      }
      elsif ($type eq "copy")
      {
	 $$value = (defined $n) ? ($n) : ($option);
      }
      elsif ($type eq "copyarray")
      {
	 @$value = @$n;
      }
      elsif ($type eq "copyhash")
      {
	 %$value = %$n;
      }
      else
      {
	 # This code looks a little more complicated than at first
	 # you might think necessary.  But the ugliness is necessary
	 # because $value might reference a scalar or an array, depending
	 # on whether $n is 1 (scalar) or not (array).  Thus, we can't
	 # just assume that either @$value or $$value is valid -- always
	 # have to check which of the two it should be.

	 if ($n > 1)
	 {
	    @$value = splice (@$arglist, 0, $n);
	    foreach $val (@$value)
	    {
	       if (! &check_value ($arg, $val, $n, $type))
	       {
		  @$arglist = @new_arglist;
		  return 0;
	       }
	    }
	 }     # if ($n > 1)
	 else
	 {
	    $$value = shift @$arglist;
	    if (! &check_value ($arg, $$value, $n, $type))
	    {
	       @$arglist = @new_arglist;
	       return 0;
	    }
	 }     # else: if ($n > 1)
      }     # else: any other type (ie. integer, string, or float)
   }     # while ($arg = shift @$arglist)

   @$arglist = @new_arglist;
   return 1;

}     # ParseArgs::Parse

1;
