# ------------------------------ MNI Header ----------------------------------
#@NAME       : file_utilities.pl
#@DESCRIPTION: Functions to manipulate/check/validate/search files and
#              directories.
#@GLOBALS    : 
#@CREATED    : 95/08/07, Greg Ward (from gpw_utilities.pl)
#@MODIFIED   : 
#@VERSION    : $Id: file_utilities.pl,v 1.8 1996/02/09 18:48:29 greg Exp $
#-----------------------------------------------------------------------------

# --------------------------------------------------------------------
# Copyright (c) 1995 Greg Ward, McConnell Brain Imaging Centre,
# Montreal Neurological Institute, McGill University.  Permission to
# use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appear in all copies.  The
# author and McGill University make no representations about the
# suitability of this software for any purpose.  It is provided "as
# is" without express or implied warranty.
# --------------------------------------------------------------------

require 5.001;

# ------------------------------ MNI Header ----------------------------------
#@NAME       : CheckOutputDirs
#@INPUT      : @dirs - list of directories to check
#@OUTPUT     : 
#@RETURNS    : 1 if all directories exist and are writeable, *or* were
#                 successfully created
#              0 otherwise
#@DESCRIPTION: Checks that each directory in a list of directories either
#              exists and is writeable; *or* that it can be created.  
#              Prints a meaningful error message and returns 0 if any fail.
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 95/5/2, Greg Ward
#@MODIFIED   : 
#-----------------------------------------------------------------------------
sub CheckOutputDirs
{
   local (@dirs) = @_;
   local ($num_err) = 0;

   foreach $dir (@dirs)
   {
      next unless $dir;		# skip blank strings
      if (-e $dir)
      {
	 if (! -d $dir)
	 {
	    warn "$dir already exists but is not a directory\n";
	    $num_err++;
	 }
	 elsif (! -w $dir)
	 {
	    warn "$dir is a directory but is not writeable\n";
	    $num_err++;
	 }
      }
      else
      {
	 if (! mkdir ($dir, 0755))
	 {
	    warn "Couldn't create $dir: $!\n";
	    $num_err++;
	 }
      }
   }
   return ($num_err == 0);
}


# ------------------------------ MNI Header ----------------------------------
#@NAME       : CheckInputDirs
#@INPUT      : @dirs - list of directories to check
#@OUTPUT     : 
#@RETURNS    : 1 if all directories exist and are readable
#              0 otherwise
#@DESCRIPTION: Checks to see if all desired input directories exist and
#              are readable.  Prints meaningful error messages and returns
#              zero if anything is wrong.
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 95/5/2, Greg Ward
#@MODIFIED   : 
#-----------------------------------------------------------------------------
sub CheckInputDirs
{
   local (@dirs) = @_;
   local ($num_err) = 0;

   foreach $dir (@dirs)
   {
      next unless $dir;		# skip blank strings
      if (-e $dir)
      {
	 if (! -d $dir)
	 {
	    warn "$dir exists but is not a directory\n";
	    $num_err++;
	 }
	 elsif (! -r $dir)
	 {
	    warn "$dir is a directory, but is not readable\n";
	    $num_err++;
	 }
      }
      else
      {
	 warn "directory $dir does not exist\n";
	 $num_err++;
      }
   }
   return ($num_err == 0);   
}


# ------------------------------ MNI Header ----------------------------------
#@NAME       : CheckDirectories
#@INPUT      : *source - reference to list of source directories
#              *dest   - reference to list of destination directories
#@OUTPUT     : 
#@RETURNS    : 
#@DESCRIPTION: Goes through the list of source directories, and makes sure
#              that each one exists; goes through list of destination
#              directories, and attempts to create those that don't exist.
#              Dies if any source directories are missing.
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 94/8/22, Greg Ward
#@MODIFIED   : 95/5/2, GW: changed to just call &CheckInputDirs and
#                          &CheckOutputDirs
#-----------------------------------------------------------------------------
sub CheckDirectories
{
   local (*source, *dest) = @_;
   local ($input_ok, $output_ok);

   $input_ok = &CheckInputDirs (@source);
   $output_ok = &CheckOutputDirs (@dest);
   return ($input_ok && $output_ok);
}


# ------------------------------ MNI Header ----------------------------------
#@NAME       : SearchDirectories
#@INPUT      : $_[0] - name of file to search for
#              $_[1] ... - list of directories to search in.  Note that
#                 the empty string should NOT be used, use "./".  
#@OUTPUT     : 
#@RETURNS    : The directory where $file was found, or undef if it wasn't.
#@DESCRIPTION: Searches a list of directories for a single file.  Note
#              that only the existence test -e is used; doesn't check
#              for readability, executability, or whatever.
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 94/9/16, Greg Ward
#@MODIFIED   : 
#-----------------------------------------------------------------------------
sub SearchDirectories
{
   local ($file, @dirs) = @_;
   local ($dir, $found);

   $found = 0;
   while ($dir = shift @dirs)
   {
      $dir .= "/" unless $dir =~ m|/$|;
      $found = (-e $dir . $file);
      last if $found;
   }

   return ($dir) if $found;
   return undef;
}


# ------------------------------ MNI Header ----------------------------------
#@NAME       : CheckFiles
#@INPUT      : @_ - a list of filenames to check for readability
#@OUTPUT     : 
#@RETURNS    : 1 if all files exist and are readable
#              0 otherwise (and prints a warning for every bad file)
#@DESCRIPTION: Makes sure that each of a list of files exists and is readable.
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : Aug 94, Greg Ward
#@MODIFIED   : 95/04/25, GW: gave it two possible error messages
#              95/10/30, GW: changed so it warns and returns 0 instead of dying
#-----------------------------------------------------------------------------
sub CheckFiles
{
   local ($num_err);

   $num_err = 0;
   foreach (@_)
   {
      if (! -e $_)
      {	
	 warn "$_ does not exist\n";
	 $num_err++;
      }
      elsif (! -r $_)
      {
	 warn "$_ not readable\n";
	 $num_err++;
      }
   }
   return ($num_err == 0);
}



# ------------------------------ MNI Header ----------------------------------
#@NAME       : &FindProgram
#@INPUT      : $program - name of program to search for
#              $path    - colon-delimited list of directories to look in
#                         (if not supplied, $ENV{"PATH"} will be used);
#                         OR a reference to a list of directories
#@OUTPUT     : 
#@RETURNS    : full path of program, or 0 if not found (also prints a 
#              warning if not found)
#@DESCRIPTION: Searches a list of directories for an executable program.
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 95/07/20, Greg Ward
#@MODIFIED   : 
#-----------------------------------------------------------------------------
sub FindProgram
{
   my ($program, $path) = @_;
   my (@path, $found, $fullpath);

   $path = $ENV{'PATH'} unless defined $path;

   if (! ref $path)
   {
      @path = split (":", $path);
   }
   elsif (ref $path eq 'ARRAY')
   {
      @path = @$path;
   }
   else 
   {
      warn "\&FindProgram: \$path must be either a scalar or ref to array\n";
      return undef;
   }

   $found = 0;
   foreach $dir (@path)
   {
      $fullpath = "${dir}/${program}";
      if (-x $fullpath && -f $fullpath)
      {
	 $found = $fullpath;
	 last;
      }
   }

   return $found if $found;
   warn "Couldn't find program $program\n";
   return 0;
}


# ------------------------------ MNI Header ----------------------------------
#@NAME       : &FindPrograms
#@INPUT      : @$programs  - list of programs to search for
#              $path       - reference to list of directories to search
#                            (OR a colon-delimited string)
#@OUTPUT     : 
#@RETURNS    : list of found programs (as full pathnames)
#              OR empty list (if there were any missing programs)
#@DESCRIPTION: Searches for each program specified in the specified search
#              path.
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : &FindProgram
#@CREATED    : 95/12/08, Greg Ward
#@MODIFIED   : 
#-----------------------------------------------------------------------------
sub FindPrograms
{
   my ($programs, $path) = @_;
   my ($missing, $found, @found);

   $missing = 0;
   foreach $prog (@$programs)
   {
      $found = &FindProgram ($prog, $path);
      $missing++ if ! $found;
      push (@found, $found);
   }
   return @found if $missing == 0;
   return ();
}


# ------------------------------ MNI Header ----------------------------------
#@NAME       : &Uncompress
#@INPUT      : $tmp_dir - directory to uncompress files to
#              @originals - list of files to consider for uncompression
#@OUTPUT     : 
#@RETURNS    : array context: @originals, changed so that any the names
#                of any files that were compressed are now decompressed
#                and in $tmp_dir
#              scalar context: first element of @originals, possibly changed
#                to decompressed version
#@DESCRIPTION: Uncompresses (if applicable) compressed files to $tmp_dir.
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 95/07/31, Greg Ward
#@MODIFIED   : 
#-----------------------------------------------------------------------------
sub Uncompress
{
   local ($tmp_dir, @originals) = @_;
   local ($uncompressed, @uncompressed);

   require "local/path_utilities.pl";	# for &ReplaceDir

   foreach $orig (@originals)
   {
      if ($orig =~ /\.(Z|gz|z)$/)
      {
	 ($uncompressed = &ReplaceDir ($tmp_dir, $orig)) =~ s/\.(Z|gz|z)$//;
	 system ("gunzip -c $orig > $uncompressed");
	 die "gunzip $orig failed\n" if ($?);
	 push (@uncompressed, $uncompressed);
      }
      else
      {
	 push (@uncompressed, $orig);
      }
   }
   return @uncompressed if wantarray;
   return $uncompressed[0];
}

1;
