# ------------------------------ MNI Header ----------------------------------
#@NAME       : jobctl
#@INPUT      : 
#@OUTPUT     : 
#@RETURNS    : 
#@DESCRIPTION: A package for doing process and job stuff -- spawning 
#              children and waiting for them (or detaching them to the
#              background), detaching ourself to the background, running
#              stuff remotely -- that sort of thing.
#@METHOD     : 
#@GLOBALS    : Expects caller to set the following flags:
#                 $Verbose      - print command name as executing?
#                 $Execute      - actually execute?
#                 $MergeErrors - if no redirection specified for stderr,
#                                 should it be merged with stdout?  (1 
#                                 means only if stdout is unchanged or
#                                 redirected to a file; 2 means always 
#                                 merge even if stdout is captured to a
#                                 variable).
#              Default on all is false, so that no commands will be
#              executed if they're not set.
#
#              Caller might also want to set:
#                 $ErrorAction
#              which is eval'd whenever a spawned command returns non-zero
#              exit status.
#
#              All of these globals can also be set by calling &SetOptions.
#@CALLS      : 
#@CREATED    : 95/3/3, Greg Ward
#@MODIFIED   : 95/11/14, GW: made into a package
#@VERSION    : $Id: job_control.pl,v 1.19 1996/02/17 21:24:36 greg Exp $
#-----------------------------------------------------------------------------

# --------------------------------------------------------------------
# Copyright (c) 1995 Greg Ward, McConnell Brain Imaging Centre,
# Montreal Neurological Institute, McGill University.  Permission to
# use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appear in all copies.  The
# author and McGill University make no representations about the
# suitability of this software for any purpose.  It is provided "as
# is" without express or implied warranty.
# --------------------------------------------------------------------

require 5.001;
require "local/misc_utilities.pl"
   unless (defined &userstamp && defined &timestamp);

package JobControl;


$ErrorAction = "return 0";
$MergeErrors = 1;
$Batch = 0;
$ClobberLogs = 0;
$LogHandle = "STDOUT";

%valid_options = ("Verbose"     => 1,
		  "Execute"     => 1,
		  "MergeErrors" => 1,
		  "ErrorAction" => 1,
		  "Batch"       => 1,
		  "ClobberLogs" => 1,
		  "LogHandle"   => 1);
		  

sub main::SetSpawnOptions
{
   local (@optval) = @_;

   ($filename,$line) = (caller)[1..2]; # for error messages

   if (@optval == 2 && $optval[0] =~ /\d+/)
   {
      warn "warning: obsolete version of \&SetSpawnOptions called from " .
	 "${filename}:${line}\n";

      ($MergeErrors, $ErrorAction) = @optval;
      return;
   }

   while (@optval)
   {
      $opt = shift @optval;
      $val = shift @optval;
      die "Unmatched option $opt at ${filename}:${line}\n" 
	 unless defined $val;
	 
      die "Unknown option $opt at ${filename}:${line}\n" unless 
	 $valid_options{$opt};

      eval "\$$opt = \$val";
   }
}


# ------------------------------ MNI Header ----------------------------------
#@NAME       : Backgroundify
#@INPUT      : $logfile - name of file to which the stdout should be redirected
#              @args    - arguments program was run with (used to generate
#                         announcement message) (just a copy of @ARGV)
#@OUTPUT     : none
#@RETURNS    : none
#@DESCRIPTION: Redirects stdout and stderr to a file (named by $logfile),
#              and detaches to background execution.  If @args is non-
#              empty, prints two announcements: to stdout, that we're
#              detaching to the background; and to the logfile, the
#              user, host, time, and full argument list (@args).
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 95/2/21 Greg Ward from nearly-identical function in do_mritopet
#@MODIFIED   : 
#@COMMENTS   : ignores the LogHandle option (for historical reasons)
#-----------------------------------------------------------------------------
sub main::Backgroundify
{
   local ($logfile, @args) = @_;
   local ($time, $user, $log_existed);

   $log_existed = (-e $logfile);
   if (!open (DUMMY, ">>$logfile"))
   {
      warn "Couldn't open log file $logfile for appending: $!\n";
      return;
   }
   close (DUMMY);

   print "Redirecting to $logfile and detaching to background\n"
      if @args && $Verbose;
   open (STDOUT, ">>$logfile") || 
      warn "Couldn't redirect stdout -- output won't be logged\n", return;
   open (STDERR, ">&STDOUT") ||
      warn "Couldn't redirect stderr -- errors won't be logged\n";

   select (STDERR); $| = 1;
   select (STDOUT); $| = 1;

   if (@args)
   {
      $user = &main::userstamp;
      $time = &main::timestamp;

      print "\n==========================================================\n"
	 if ($log_existed);
      print "Log file opened by $user at $time for command:\n";
      print $0 . " " . join (" ", @args) . "\n\n";
   }

   fork && exit;
}




# ------------------------------ MNI Header ----------------------------------
#@NAME       : &check_return_status
#@INPUT      : 
#@OUTPUT     : 
#@RETURNS    : 
#@DESCRIPTION: Checks the return status of a program.  If non-zero,
#              prints a warning and eval's $ErrorAction (if it
#              is defined).
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 96/01/17, GPW (from code in &Spawn)
#@MODIFIED   : 
#-----------------------------------------------------------------------------
sub check_return_status
{
   my ($status, $program) = @_;

   if ($status)
   {
      warn "$program returned non-zero status code\n";
      if ($ErrorAction)
      {
	 my ($i, $this_pkg, @caller);

	 $i = 0;
	 $this_pkg = (caller(0))[0];
	 while (@caller = caller($i++))
	 {
	    $up_pkg = $caller[0];
	    last if $up_pkg ne $this_pkg;
	 }

	 eval "package $up_pkg; $ErrorAction";
	 die $@ if $@;
      }
   }
}



# ------------------------------ MNI Header ----------------------------------
#@NAME       : &Spawn
#@INPUT      : $command - the command (and all arguments) to run
#              $stdout  - tells what to do with the standard output.
#                         Can be either a filename (redirect to that
#                         file), "-" (capture to a variable, which is
#                         returned by &Spawn), or empty (leave it alone).
#                         If $stdout is a filename, it will be either
#                         overwritten or appended depending on the 
#                         ClobberLogs option.
#              $stderr  - identical to $stdout.  Note that you
#                         can merge the standard error into the standard
#                         output by using ">&STDOUT" for $stderr (this is
#                         a standard Perl trick).
#              $stdin   - doesn't do anything (yet) (and probably never will)
#@OUTPUT     : 
#@RETURNS    : $status - exit status of the child process
#              $output - (if $stdout was "$") - standard output of the 
#                        child process
#@DESCRIPTION: Runs a child process, optionally capturing its standard
#              output and/or standard error to a file or a Perl variable.
#              To capture either stdout or stderr to a filename, just pass
#              filenames as you would with the open function (eg. ">foo.log")
#              as $stdout and/or $stdin.  You can merge stderr with stdout
#              either by setting $stderr to ">&STDOUT", or by leaving
#              $stderr empty (or undefined) and setting the global variable
#              $MergeErrors to 1 or 2.
#
#              To capture to a variable, simply evaluate &Spawn in an array
#              context, eg. ($status, $output, $error) = &Spawn (...).  By 
#              default, this will make stdout and stderr go to separate
#              variables; you can counteract this (as above) by setting
#              $stderr to ">&STDOUT" or by setting $MergeErrors to 2.
#@METHOD     : 
#@GLOBALS    : $Verbose, $Execute, $MergeErrors, $ErrorAction
#@CALLS      : 
#@CREATED    : 95/3/3, Greg Ward (didn't really get anywhere)
#@MODIFIED   : 95/5/16, GW - finished it (well, mostly)
#              95/6/20, GW - added code for capturing stdout and stderr
#-----------------------------------------------------------------------------
sub main::Spawn
{
   local ($command, $stdout, $stderr, $stdin) = @_;
   local ($program);
   local ($pid, $status, $output, $error);

   ($Verbose = $main::Verbose, 
    warn "Spawn: warning: silently inheriting main::Verbose\n")
      unless defined $Verbose;
   ($Execute = $main::Execute,
    warn "Spawn: warning: silently inheriting main::Execute\n")
      unless defined $Execute;

   $stdout = "" if (! defined $stdout);
   $stderr = "" if (! defined $stderr);
   $stdin = "" if (! defined $stdin);

   # Check for unimplemented features

   die "&Spawn: sorry, can't redirect stdin at all (yet)\n"
      if ($stdin ne "");

   # If Batch mode is on, submit this command and be done with it

   if ($Batch)
   {
      require "local/batch.pl";

      $stdout = "" if $stdout eq "-"; # can't capture
      $stderr = "" if $stderr eq "-";
      &Batch::QueueCommand ($command, "", $stdout, $stderr, $MergeErrors);
      return
   }

   # Get the first word (i.e. the program to run) from the command string

   ($program) = split (/\s+/, $command);

   # Capture both stdout and stderr to variables if we're evaluated
   # in an array context and a redirection destination (ie. file) 
   # hasn't already been specified for them.

   $stdout = "-" if ($stdout eq "" && wantarray);
   $stderr = "-" if ($stderr eq "" && wantarray);

   # If nothing was explicitly said about what to do with $stderr,
   # follow the advice of $MergeErrors (i.e. if capturing to a variable,
   # then merge stderr only if $MergeErrors is 2; if not capturing, then
   # merge by default if $MergeErrors is 1).

   if ($stderr eq "")
   {
      $stderr = ">&STDOUT" if $MergeErrors >= 1;
   }
   elsif ($stderr eq "-")
   {
      $stderr = ">&STDOUT" if $MergeErrors >= 2;
   }

   # Figure out how to open files (> to overwrite or >> to append),
   # and hard-code that into $stdout and $stderr if they don't
   # already have such a code in them.

   $open_prefix = ($ClobberLogs) ? ">" : ">>";
   $stdout = $open_prefix . $stdout if ($stdout && $stdout !~ /^[>-]/);
   $stderr = $open_prefix . $stderr if ($stderr && $stderr !~ /^[>-]/);

   # Report the user, host, date, time, and command to be run

   printf $LogHandle "%s %s %s\n", &main::userstamp, &main::timestamp, $command
      if $Verbose;
   return 0 unless $Execute;

   # Do things rather differently depending on whether we are to
   # capture stdout.  In the capture-to-variable case, we will
   # fork-and-exec the desired command, and slurp its output via a
   # pipe; stderr will be captured via a temporary file (if stderr is
   # supposed to be captured - it might be redirected or left alone).
   # (Yeah, I know stdout could just as well have been captured the
   # same way.  However, I was 1) curious about doing it via a pipe,
   # and 2) concerned about efficiency -- error output is usually
   # pretty small so going via disk won't kill anyone, but stdout
   # could be big.)  I didn't do stderr via a pipe because I don't
   # think you can have *two* pipes (one for stdout and one for
   # stderr) connecting the two processes.  Perhaps it could be done
   # with a named pipe... ah well, another day...

   if ($stdout eq "-")		# capturing to a variable
   { 
      # Run the command through a pipe.  In the child, stdout
      # automagically goes to the pipe; in the parent, we will slurp
      # the child's output from the pipe.

      $pid = open (PIPE, "-|");
      die "Spawn: failed to start child process: $!\n" unless defined $pid;
      if ($pid == 0)		# in the child?  
      {
	 # We're in the child so don't do much: do error redirection if
	 # any (including merging into stdout - this is handled above),
	 # and then exec.

	 if ($stderr eq "-")	# caller wants it in a variable -- we'll
	 {			# be sneaky and put it in a file for later
	    open (STDERR, ">/tmp/error$$.log");	# retrieval 
	 }
	 elsif ($stderr ne "")	# caller wants it in a file (including
	 {			# possibly &STDOUT), so just put it there
	    open (STDERR, $stderr);
	 }

	 if (! exec ($command))
	 {
	    die "Unable to execute $program: $!\n";
	 }
      }
      else			# $pid != 0 => in the parent
      {
	 # We're in the parent, so we read in the child's stdout
	 # (and stderr too if it was captured).

	 $output = join ("", <PIPE>);
	 close (PIPE);
	 $status = $? >> 8;

	 if ($stderr eq "-")	# was stderr captured?
	 {
	    open (ERROR, "</tmp/error${pid}.log");
	    $error = join ("", <ERROR>);
	    close (ERROR);
	    unlink ("/tmp/error${pid}.log");
	 }

	 &check_return_status ($status, $program);
	 return ($status, $output, $error);
      }
   }

   # If capturing to file(s), do it right -- fork, do redirection 
   # in the child process, and exec the command; in the parent, just
   # wait for the child to finish

   $pid = fork;
   die "Spawn: failed to start child process: $!\n" unless defined $pid;
   if ($pid == 0)		# we're in the child process
   {
      # Redirect stdout and/or stderr if the caller specified file 
      # destination(s)

      if ($stdout ne "") 
      {
	 open (STDOUT, $stdout) 
	    || die "Unable to redirect stdout to $stdout: $!\n";
      }
      if ($stderr eq "-") 
         { open (STDERR, ">/tmp/error$$.log"); }
      elsif ($stderr ne "")
         { open (STDERR, $stderr); }

      if (! exec ($command))
      {
	 die "Unable to execute $program: $!\n";
      }
   }
   else				# we're still in the parent
   {
      # block and wait for child

      if (waitpid ($pid, 0) == -1)
      {
	 die "&Spawn: no child processes even though I just forked!\n";
      }
      
      $status = $? >> 8;

      if ($stderr eq "-")	# was stderr captured?
      {
	 open (ERROR, "</tmp/error${pid}.log");
	 $error = join ("", <ERROR>);
	 close (ERROR);
	 unlink ("/tmp/error${pid}.log");
      }

      &check_return_status ($status, $program);

      return ($status, $output, $error) if wantarray;
      return $status;
   }      
}


sub main::Obituary		# use this for long-running programs that
{				# send mail when done or crashed
   my ($command, $status, $error) = @_;
   my ($program, $text);

   $Notify = $main::Notify unless defined $Notify;

   return unless ($Notify && $Execute);

   if ($command)
   {
      $program = (split (/\s+/, $command))[0];
      $text = "$main::ProgramName crashed while executing the command:";
   }
   else
   {
      $text = "$main::ProgramName crashed";
   }
   open (MAIL, "|mail $Notify");

   print MAIL <<EOM;
Subject: $main::ProgramName crashed!

Dear $Notify,

$text

EOM
   print MAIL "> $command\n" if ($command);
   print MAIL "Exit status: $status\n" if (defined $status);
   if (defined $error)
   {
      print MAIL "Error message(s): \n$error\n";
   }
   else
   {
      print MAIL "You'll have to check the log file to see what exactly went wrong.\n";
   }

   close (MAIL);
#   &Fatal ("crashed while running $program");
    if ($command)
       { &main::Fatal ("crashed while running $program"); }
    else
       { &main::Fatal ("crashed"); }
}



1;
