# ------------------------------ MNI Header ----------------------------------
#@NAME       : minc_utilities.pl
#@INPUT      : 
#@OUTPUT     : 
#@RETURNS    : 
#@DESCRIPTION: Interfaces to a various MINC utilities for common
#              image- and volume-processing tasks.  Uses &Spawn to run
#              the commands, so be sure you have set the error action
#              to something reasonable (using &SetSpawnOptions).
#              Also, requires the following global variables to give
#              the full location and standard options for the
#              programs:
#                 $MincInfo
#                 $VolumeStats
#                 $VolumeCOG
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : Greg Ward, July 95
#@MODIFIED   : considerably
#@VERSION    : $Id: minc_utilities.pl,v 1.9 1996/02/12 19:58:18 greg Exp $
#-----------------------------------------------------------------------------

# --------------------------------------------------------------------
# Copyright (c) 1995 Greg Ward, McConnell Brain Imaging Centre,
# Montreal Neurological Institute, McGill University.  Permission to
# use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appear in all copies.  The
# author and McGill University make no representations about the
# suitability of this software for any purpose.  It is provided "as
# is" without express or implied warranty.
# --------------------------------------------------------------------

require 5.001;
require "local/job_control.pl" unless defined &Spawn;

sub volume_minmax
{
   my ($volume) = @_;
   my ($status, $volmax, $volmin);
   die "\$MincInfo undefined" unless defined $MincInfo;

   ($status, $volmin) = 
      &Spawn ("$MincInfo -varvalue image-min $volume | sort -n | head -1");
   ($status, $volmax) = 
      &Spawn ("$MincInfo -varvalue image-max $volume | sort -n | tail -1");
   chop $volmin;
   chop $volmax;
   $Execute ? ($volmin, $volmax) : (0,0);
}


sub volume_threshold
{
   my ($volmin, $volmax, $percent) = @_;
   ($volmax - $volmin) * $percent + $volmin;
}


sub auto_threshold
{
   my ($volume) = @_;
   die "\$VolumeStats undefined" unless defined $VolumeStats;

   ($status, $threshold) = 
      &Spawn ("$VolumeStats -biModalT -quiet $volume");
   chop $threshold;
   ($Execute) ? $threshold : 0;
}


sub volume_cog
{
   my ($volume) = @_;
   my ($cog, @cog);
   die "\$VolumeCOG undefined" unless defined $VolumeCOG;

   ($status, $cog) = &Spawn ("$VolumeCOG $volume | tail -1");
   chop $cog;
   ($Execute) ? $cog : "0 0 0";
}


# ------------------------------ MNI Header ----------------------------------
#@NAME       : volume_params
#@INPUT      : $volume - file to query
#              $start  - reference to list to fill with starts
#              $step   - reference to list to fill with step sizes
#              $length - reference to list to fill with dimension lengths
#              $dircos - reference to list to fill with direction cosines
#              (If any of these "references" are just false values, then
#              they won't be followed -- that way you can specify "undef"
#              for values you're not interested in.)
#@OUTPUT     : 
#@RETURNS    : 
#@DESCRIPTION: Gets the x, y, and z starts, steps, and lengths for a
#              MINC volume.
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 95/04/21, Greg Ward
#@MODIFIED   : 95/08/04, GW: simplified to handle negative steps and extents
#              95/08/16, GW: changed to GetVolumeParams and made so it returns
#                            all three interesting spatial parameters
#              95/08/30, GW: renamed to volume_params and moved from autocrop
#                            to minc_utilities.pl
#              96/02/12, GW: converted to use Perl 5 references
#                            added direction cosines stuff
#-----------------------------------------------------------------------------
sub volume_params
{
   my ($volume, $start, $step, $length, $dircos) = @_;
   my ($cmd, $output, @output);

   $cmd = "mincinfo $volume -error_string 0 " .
          "-attval xspace:start -attval yspace:start -attval zspace:start " .
          "-attval xspace:step -attval yspace:step -attval zspace:step " .
	  "-dimlength xspace -dimlength yspace -dimlength zspace " .
	  "-attval xspace:direction_cosines " .
	  "-attval yspace:direction_cosines " .
	  "-attval zspace:direction_cosines";
   $output = `$cmd`;
   die "Error executing mincinfo on $volume\n" if $?;
   @output = split ("\n", $output);
   grep (s/(^\s+)|(\s+$)//g, @output); # strip leading and trailing whitespace

   @$start = @output[0..2] if $start;
   @$step = @output[3..5] if $step;
   @$length = @output[6..8] if $length;
   if ($dircos)
   {
      @$dircos = (1,0,0, 0,1,0, 0,0,1);
      for $i (0,1,2)
      {
	 my ($base) = $i * 3;

	 @$dircos[$base .. $base+2] = split (/\s+/, $output[9+$i])
	    unless ($output[9+$i] eq "0");
      }
   }
}


sub get_history
{
   my ($volume) = @_;
   my ($history, @history);

   $history = `mincinfo -error_string "" -attvalue :history $volume`;
   die "Error executing mincinfo on $volume\n" if $?;
   @history = split ("\n", $history);
   pop (@history) unless $history[-1];
   @history;
}


sub put_history
{
   my ($volume, @history) = @_;
   my ($history);

   $history = join ("\n", @history);
   system ("minc_modify_header $volume -sinsert :history=\"$history\"");
   die "Error modifying header of $volume\n" if $?;
}


1;
