# ------------------------------ MNI Header ----------------------------------
#@NAME       : path_utilities.pl
#@DESCRIPTION: Functions for recognizing, parsing, and tweaking Unix
#              filenames and paths:
#                 &ReplaceDir
#                 &ReplaceExt
#                 &SplitPath
#@GLOBALS    : 
#@CREATED    : 95/08/07, Greg Ward (from gpw_utilities.pl)
#@MODIFIED   : 
#@VERSION    : $Id: path_utilities.pl,v 1.5 1995/12/04 21:27:00 greg Exp $
#-----------------------------------------------------------------------------

# --------------------------------------------------------------------
# Copyright (c) 1995 Greg Ward, McConnell Brain Imaging Centre,
# Montreal Neurological Institute, McGill University.  Permission to
# use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appear in all copies.  The
# author and McGill University make no representations about the
# suitability of this software for any purpose.  It is provided "as
# is" without express or implied warranty.
# --------------------------------------------------------------------


# ------------------------------ MNI Header ----------------------------------
#@NAME       : ReplaceDir
#@INPUT      : $newpath - directory to replace existing directories with
#              @files   - list of files to have directories replaced
#@OUTPUT     : 
#@RETURNS    : 
#@DESCRIPTION: Replaces the directory component of a list of pathnames.
#              Returns the list of files with substitutions performed.
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 95/5/4, Greg Ward
#@MODIFIED   : 95/05/23, GW: renamed to &ReplaceDir
#-----------------------------------------------------------------------------
sub ReplaceDir
{
   local ($newpath, @files) = @_;

   $newpath = "." if $newpath eq "";
   $newpath =~ s|([^/])$|$1/|;	# add a trailing slash if necessary
   foreach $file (@files)
   {
      $file =~ s|.*/||;		# strip off any existing directory
      $file = $newpath . $file;
   }
   return @files if wantarray;
   return $files[0];
}


# ------------------------------ MNI Header ----------------------------------
#@NAME       : ReplaceExt
#@INPUT      : $newext  - extension to replace existing extensions with
#              @files   - list of files to have extensions replaced
#@OUTPUT     : 
#@RETURNS    : 
#@DESCRIPTION: Replaces the final extension (whatever follows the final dot)
#              of a list of pathnames.  Returns the list of files with
#              substitutions performed in array context, or the first filename
#              from the list in a scalar context.
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 95/05/23, Greg Ward (from &ReplaceDir)
#@MODIFIED   : 
#-----------------------------------------------------------------------------
sub ReplaceExt
{
   local ($newext, @files) = @_;

   foreach $file (@files)
   {
      $file =~ s/\.[^\.]*/\.$newext/; # replace existing extension
   }
   return @files if wantarray;
   return $files[0];
}
      


# ------------------------------ MNI Header ----------------------------------
#@NAME       : &SplitPath
#@INPUT      : $input - a Unix path specifiction (optional directory + file)
#              $ext_opt - specifies how to deal with file extensions:
#                         if "none", extension is ignored and returned as
#                           part of the base filename
#                         if "first", the *first* dot in a filename denotes
#                           the extension, eg ".mnc.gz" would be an extension
#                         if "last", the *last* dot denotes the extension,
#                           eg. just ".gz" would be the extension
#                         the default is "first"
#@OUTPUT     : 
#@RETURNS    : array: ($directory, $basename, $extension)
#@DESCRIPTION: Splits a Unix path specification into directory, base file 
#              name, and extension.  (The extension is chosen based on
#              either the first or last dot in the filename, depending
#              on the $ext_opt argument; by default, it splits on the 
#              first dot in the filename.)
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 95/05/10, Greg Ward - taken from mritotal and modified
#@MODIFIED   : 95/08/10, GW: added $ext_opt option to handle splitting off
#                            the extension in different ways
#-----------------------------------------------------------------------------
sub SplitPath
{
    local($input, $ext_opt) = @_;
    local ($dir, $base, $ext);

    $ext_opt = "first" unless defined $ext_opt;

    # If filename has no extension, don't try to act as though it does
    # (both "last" and "first" options assume there is an extension

#    $ext_opt = "none" if $input !~ m+/?[^/]*\.+;

    if ($ext_opt eq "none")
    {
       ($dir, $base) = $input =~ m+^(.*/)?([^/]*)$+;
    } 
    elsif ($ext_opt eq "first")
    {
       ($dir, $base, $ext) = $input =~ m+^(.*/)?([^/\.]*)(\..*)?$+;
    }
    elsif ($ext_opt eq "last")
    {
       ($dir, $base, $ext) = $input =~ m+^(.*/)?([^/]*)(\.[^\/]*)$+;
    }
    else
    {
       die "SplitPath: unknown extension option \"$ext_opt\"\n";
    }

    if (!$dir)                  # no dir found? then use current dir
    {
       $dir = ".";
    }

    $dir =~ s/\/$//;		# strip off trailing slash

    ($dir, $base, $ext);
}


# ------------------------------ MNI Header ----------------------------------
#@NAME       : &MergePaths
#@INPUT      : $path      - ref to a list of directories
#              $additions - ref to a list of directories to add to $path
#@OUTPUT     : @$path     - updated so elements from @$additions are added
#                           only if they weren't already in @$path
#@RETURNS    : 
#@DESCRIPTION: 
#@METHOD     : 
#@GLOBALS    : 
#@CALLS      : 
#@CREATED    : 95/12/04 GPW
#@MODIFIED   : 
#-----------------------------------------------------------------------------
sub MergePaths
{
   my ($path, $additions) = @_;

   foreach $add (@$additions)
   {
      push (@$path, $add) unless grep ($add eq $_, @$path);
   }
}

1;
