# ------------------------------ MNI Header ----------------------------------
#@NAME       : startup.pl
#@INPUT      : 
#@OUTPUT     : 
#@RETURNS    : 
#@DESCRIPTION: Code that should be executed in all "serious" Perl scripts.
#              Sets up signal handlers, names temp directory, cleanup 
#              routines, ....
#@METHOD     : 
#@GLOBALS    : The following variables are put into main's namespace after
#              you call &Startup:
#                $ProgramPath, $ProgramName - the two components of $0
#                $TmpDir - a (hopefully) uniquely named temporary directory
#                $StartDir - the working directory at startup
#                $StartDirName - the final component of the directory from
#                            which we were run (eg. if pwd = 
#                            "/usr/local/poobah" on startup, then
#                            $StartDirName will be "poobah")
#                @DefaultArgs - a basic argument table (for use with my
#                               ParseArgs package) for setting common
#                               globals ($Verbose, $Execute, $Debug, $Clobber,
#                               $Debug, $TmpDir, and $KeepTmp) via common
#                               options (-verbose, -execute, -debug, etc)
#
#              Also, &Startup modifies the %SIG builtin hash.
#
#              The following should be in main's namespace when you call
#              &Cleanup (and note that &Cleanup can be called at any time,
#              since the signal handler eventually ends up there):
#                $KeepTmp - set to true to not delete the temporary 
#                           directory and everything in it.
#                $Notify  - set the the email address of the user to notify
#                           in case the program crashes.  (Well, really
#                           in case &Obituary is called, which is usually
#                           when the program crashes.)  Set to the empty
#                           string (or any false value) to disable
#                           email notification.
#                $Execute - must also be set `true' to enable email
#                           notification (this is just so the user doesn't
#                           get spurious email for test runs).
#@CALLS      : 
#@CREATED    : 95/05/16, Greg Ward: from code in mritotal, do_mritopet, and
#                        autocrop
#@MODIFIED   : 95/07/07, GW: renamed from gpw_common.pl to startup.pl,
#                        moved the actual startup code into &Startup,
#                        added tons of comments
#              95/08/22, GW: removed $PrintTimes from list of needed globals
#                        (now it's just a parameter to &Cleanup); cosmetic
#                        and doc changes
#@COMMENTS   : doesn't handle argument parsing, since that's currently 
#              dependent on the &ParseArgs routine in every program (arg!)
#@VERSION    : $Id: startup.pl,v 1.17 1996/01/16 21:58:25 greg Exp $
#----------------------------------------------------------------------------

# --------------------------------------------------------------------
# Copyright (c) 1995 Greg Ward, McConnell Brain Imaging Centre,
# Montreal Neurological Institute, McGill University.  Permission to
# use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appear in all copies.  The
# author and McGill University make no representations about the
# suitability of this software for any purpose.  It is provided "as
# is" without express or implied warranty.
# --------------------------------------------------------------------

# To use startup.pl, you should (obviously) have `require
# "startup.pl"' in your code.  Then, before any serious work
# (eg. creating directories, processing files, etc.) is done, call
# &Startup.  This will put a number of global variables into the
# namespace of the calling package (usually main); see above for
# complete details.  Also, it will setup a signal handler (for the
# most common signals) that will 1) print a useful error message, 2)
# cleanup $TmpDir, and 3) quit with non-zero exit status.
#
# Then, once serious processing has begun (ie. if there might be files
# in $TmpDir), you should call &Fatal instead of die for any errors;
# this will properly clean up and leave a non-zero exit status for the
# caller.)  When everything finishes normally, call &Cleanup as one of
# the last things you do.  (It doesn't have to be the last, as it
# doesn't exit -- it just rm -rf's the temporary directory.)  Also,
# you can optionally make &Cleanup print out the running time spent by
# your script and its children -- just pass in a true value for its
# first argument.

# ------------------------------------------------------------------ #

require 5.001;


# ------------------------------------------------------------------ #
#   public subroutines                                               #
# ------------------------------------------------------------------ #

$KeepTmp = 1;			# so we don't nuke current dir if
				# caller forgets to call &Startup

#
# &Startup
#
# call this when your program starts -- this defines several globals that
# are almost always useful, as well as defining a signal handler for
# the more common signals.  Also unbuffers stdout and stderr, which 
# is nice for "tail -f"'ing log files, and that sort of thing.
#
sub Startup
{
   ($ProgramPath,$ProgramName) = $0 =~ /(.*\/)?([^\/]*)/;

   local ($basetmp) = (defined ($ENV{"TMPDIR"}) ? $ENV{"TMPDIR"} : "/tmp");
   $TmpDir = ($basetmp . "/${ProgramName}_$$");
   $KeepTmp = 0;
   
   $SIG{'INT'} = $SIG{'TERM'} = $SIG{'QUIT'} = $SIG{'HUP'} = 'catch_signal';
   
   chop ($StartDir = `pwd`);
   ($StartDirName) = $StartDir =~ /.*\/([^\/]+)/;

   select (STDERR); $| = 1;
   select (STDOUT); $| = 1;
   
   @Startup::start_times = times;

   @DefaultArgs =
      (["Basic behaviour options", "section"],
       ["-verbose|-quiet", "boolean", 0, \$Verbose, 
	"print status information and command lines of subprograms " .
	"[default; opposite is -quiet]" ],
       ["-execute|-noexecute", "boolean", 0, \$Execute, 
	"actually execute planned commands [default]"],
       ["-clobber|-noclobber", "boolean", 0, \$Clobber,
	"blithely overwrite files (and make subprograms do as well) " .
	"[default: -noclobber]"],
       ["-debug|-nodebug", "boolean", 0, \$Debug,
	"spew lots of debugging info (and make subprograms do so as well) " .
	"[default: -nodebug]"],
       ["-tmpdir", "string", 1, \$TmpDir,
	"set the temporary working directory"],
       ["-keeptmp|-nokeeptmp", "boolean", 0, \$KeepTmp,
	"don't delete temporary files when finished [default: -nokeeptmp]"]);
}


#
# &SelfAnnounce
#
# prints the user, host, time, and full command line (as supplied in @args)
# to stdout.  Useful for later figuring out what happened from a log file.
#
sub SelfAnnounce
{
   local (@args) = @_;

   printf "Starting command by %s at %s:\n", &userstamp, &timestamp;
   print "> $0 " . join(" ", @args) . "\n\n";
}


#
# &Cleanup
#
# should be called when your program exits, either normally or
# abnormally.  Takes one argument ($print_times - toggles whether we
# should print the process' CPU time); assumes one globals ($KeepTmp -
# toggles whether we should delete temporary files).  Does not exit
# the program -- just (optionally) prints the running time, and
# (optionally) nukes everything in $TmpDir.
#
sub Cleanup                # call this when your program finishes normally
{
   local ($print_times) = @_;
   if ($print_times)
   {
      local (@stop_times, $i, $user, $system);

      @stop_times = times;
      foreach $i (0 .. 3)
      { 
	 $elapsed[$i] = $stop_times[$i] - $Startup::start_times[$i];
      }
      $user = $elapsed[0] + $elapsed[2];
      $system = $elapsed[1] + $elapsed[3];
      print "Elapsed time in ${ProgramName}:\n";
      printf "%g sec (user) + %g sec (system) = %g sec (total)\n", 
	      $user, $system, $user+$system;
   }

   system ("rm -rf $TmpDir") unless ($KeepTmp || $TmpDir eq "");
}


# 
# &Fatal
# 
# should be called in case of emergency, ie. in place of die after
# your program starts doing "serious processing".  Has one optional
# argument, the error message to be printed.
#
sub Fatal
{
   local ($message) = @_;
   print STDERR "$ProgramName: $message\n" if defined $message;
   &cleanup_and_die ();
}


# ------------------------------------------------------------------ #
#   private subroutines and global variables                         #
# ------------------------------------------------------------------ #

package Startup;


# Most of the important signals (taken from <sys/signal.h> under IRIX 5.2) 

%signals = ("HUP", "hung-up", 
	    "INT", "interrupted", 
	    "QUIT", "quit",
	    "ILL", "illegal instruction",
	    "TRAP", "trace trap",
	    "ABRT", "aborted",
	    "EMT", "EMT instruction",
	    "FPE", "floating-point exception",
	    "KILL", "killed",
	    "BUS", "bus error",
	    "SEGV", "segmentation violation",
	    "SYS", "bad argument to system call",
	    "PIPE", "write on a pipe with no one to read it",
	    "ALRM", "alarm clock",
	    "TERM", "terminated");



sub main::cleanup_and_die	# you probably shouldn't call this 
{				# (because it dies silently)
   &main'Cleanup ();
   exit 1;
}

sub main::catch_signal		# %SIG points to this - you shouldn't call it
{
   local ($sig) = @_;
   &main'Fatal ("$signals{$sig}");
}
